import configparser
import functools
import os
import sys
from pathlib import Path

import click
from loguru import logger

CONFIG_PATH = os.path.expanduser("~/.qulab.ini")
ENV_PREFIX = "QULAB_"


def _get_config_value(option_name, type_cast=str, command_name=None):
    """支持命令专属配置的优先级获取"""
    # 构造环境变量名
    if command_name:
        env_var = f"{ENV_PREFIX}{command_name.upper()}_{option_name.upper()}"
        config_section = command_name
    else:
        env_var = f"{ENV_PREFIX}{option_name.upper()}"
        config_section = "common"

    # 1. 检查环境变量
    if env_value := os.getenv(env_var):
        if type_cast is bool:
            return env_value.lower() in ("true", "1", "yes")
        if "path" in option_name or issubclass(type_cast, Path):
            return os.path.expanduser(env_value)
        return type_cast(env_value)

    # 2. 检查配置文件
    config = configparser.ConfigParser()
    # 先加载默认配置防止段不存在
    config.read_dict({config_section: {}})
    if Path(CONFIG_PATH).exists():
        config.read(CONFIG_PATH)

    # 从对应配置段读取
    if config.has_section(config_section):
        if config_value := config.get(config_section,
                                      option_name,
                                      fallback=None):
            if type_cast is bool:
                return config_value.lower() in ("true", "1", "yes")
            if "path" in option_name or issubclass(type_cast, Path):
                return os.path.expanduser(config_value)
            return type_cast(config_value)

    return None  # 交给 Click 处理默认值


def get_config_value(option_name, type_cast=str, command_name=None):
    value = _get_config_value(option_name, type_cast, command_name)
    if value is None and command_name is not None:
        return _get_config_value(option_name, type_cast)
    return value


def log_options(func):
    """通用配置装饰器（所有命令共用）"""

    @click.option("--debug",
                  is_flag=True,
                  default=get_config_value("debug", bool),
                  help=f"Enable debug mode")
    @click.option("--log",
                  type=click.Path(),
                  default=lambda: get_config_value("log", Path),
                  help=f"Log file path")
    @click.option("--quiet",
                  is_flag=True,
                  default=get_config_value("quiet", bool),
                  help=f"Disable log output")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        debug = bool(kwargs.pop("debug"))
        log = kwargs.pop("log")
        quiet = bool(kwargs.pop("quiet"))

        if debug:
            log_level = "DEBUG"
        else:
            log_level = "INFO"

        handlers = []
        if log is not None:
            handlers.append(dict(sink=log, level=log_level))
        if not quiet or debug:
            handlers.append(dict(sink=sys.stderr, level=log_level))

        logger.configure(handlers=handlers)

        return func(*args, **kwargs)

    return wrapper



import itertools

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import PathPatch
from matplotlib.path import Path


def cycle(x0, y0, r, a=0, b=2 * np.pi):
    t = np.linspace(a, b, 101)
    x = r * np.cos(t) + x0
    y = r * np.sin(t) + y0
    return x, y


def plotSquare(xc, yc, xs, ys, radius=0, color='C0', ax=None):
    ax = plt.gca() if ax is None else ax

    l, r, b, u = xc - xs / 2, xc + xs / 2, yc - ys / 2, yc + ys / 2
    x = np.linspace(l + radius, r - radius, 2)
    y = np.linspace(b + radius, u - radius, 2)

    ax.plot(x, [u, u], color=color)
    ax.plot(x, [b, b], color=color)
    ax.plot([l, l], y, color=color)
    ax.plot([r, r], y, color=color)

    xx, yy = cycle(x[-1], y[-1], radius, 0, np.pi / 2)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[0], y[-1], radius, np.pi / 2, np.pi)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[0], y[0], radius, np.pi, np.pi * 3 / 2)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[-1], y[0], radius, np.pi * 3 / 2, np.pi * 2)
    ax.plot(xx, yy, color=color)


def plotMesure(x0=0, y0=0, size=1, color='C0', ax=None):

    ax = plt.gca() if ax is None else ax

    x, y = cycle(x0, y0 - 0.37 * size, 0.57 * size, np.pi / 4, np.pi * 3 / 4)
    ax.plot(x, y, color=color)

    plotSquare(x0, y0, size, 0.8 * size, radius=0.2 * size, color=color, ax=ax)

    angle = np.pi / 6

    ax.arrow(x0,
             y0 - 0.25 * size,
             0.55 * size * np.sin(angle),
             0.55 * size * np.cos(angle),
             head_width=0.07 * size,
             head_length=0.1 * size,
             color=color)


def plot_seq(ax, waves, measure=[1], gap=4, maxTime=20, xlim=None):
    t = np.linspace(0, maxTime, 1001)
    tt = np.linspace(0, maxTime - gap, 1001)
    styles = ['-', '--', ':', '-.']
    colors = ['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']
    for i, (wavgroup, color) in enumerate(zip(waves, itertools.cycle(colors))):
        if i in measure:
            time = tt
        else:
            time = t
        if not isinstance(wavgroup, tuple):
            wavgroup = (wavgroup, )

        for wav, style in zip(wavgroup, itertools.cycle(styles)):
            ax.plot(time, wav(time) - gap * i, style, color=color)
            if i in measure:
                plotMesure(maxTime - gap / 2,
                           -gap * i,
                           gap,
                           color='black',
                           ax=ax)

    ax.axis('equal')

    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    if xlim is not None:
        ax.set_xlim(xlim)


def make_path(fx, fy, dfx, dfy, t):

    def p(fx, fy, dfx, dfy, t0, t1):
        x1, x2 = fx(t0), fx(t1)
        y1, y2 = fy(t0), fy(t1)
        dx1, dx2 = dfx(t0), dfx(t1)
        dy1, dy2 = dfy(t0), dfy(t1)

        xp = (dy1 * dx2 * x1 - dy2 * dx1 * x2 - dx1 * dx2 *
              (y1 - y2)) / (dy1 * dx2 - dy2 * dx1)

        yp = (dy1 * dx2 * y2 - dy2 * dx1 * y1 + dy1 * dy2 *
              (x1 - x2)) / (dy1 * dx2 - dy2 * dx1)

        return xp, yp

    path = [(fx(t[0]), fy(t[0]))]
    codes = [Path.MOVETO]

    for i in range(1, len(t)):
        x, y = p(fx, fy, dfx, dfy, t[i - 1], t[i])
        path.append((x, y))
        codes.append(Path.CURVE3)
        path.append((fx(t[i]), fy(t[i])))
        codes.append(Path.CURVE3)
    return path, codes


def plot_square(ax,
                x=0,
                y=0,
                width=1,
                hight=1,
                r=0.2,
                ls='-',
                lw=2,
                fc='none',
                ec='black'):

    path = np.array([(0, 1), (1 - 2 * r, 1), (1, 1), (1, 1 - 2 * r),
                     (1, -1 + 2 * r), (1, -1), (1 - 2 * r, -1),
                     (-1 + 2 * r, -1), (-1, -1), (-1, -1 + 2 * r),
                     (-1, 1 - 2 * r), (-1, 1), (-1 + 2 * r, 1), (0, 1)])
    path[:, 0] = path[:, 0] * width / 2 + x
    path[:, 1] = path[:, 1] * hight / 2 + y
    codes = [
        Path.MOVETO,
        Path.LINETO,
        Path.CURVE3,
        Path.CURVE3,
        Path.LINETO,
        Path.CURVE3,
        Path.CURVE3,
        Path.LINETO,
        Path.CURVE3,
        Path.CURVE3,
        Path.LINETO,
        Path.CURVE3,
        Path.CURVE3,
        Path.CLOSEPOLY,
    ]
    pp1 = PathPatch(Path(path, codes),
                    ls=ls,
                    lw=lw,
                    fc=fc,
                    ec=ec,
                    transform=ax.transData)

    ax.add_patch(pp1)


def plot_measure(ax,
                 x=0,
                 y=0,
                 size=1,
                 r=0.2,
                 ls='-',
                 lw=2,
                 fc='none',
                 ec='black'):
    width = size
    hight = 0.8 * size
    plot_square(ax, x, y, width, hight, r, ls, lw, fc, ec)

    fx = lambda t: 0.5 * width * np.cos(t) + x
    fy = lambda t: 0.5 * hight * np.sin(t) + y - 0.2 * hight
    dfx = lambda t: -0.5 * width * np.sin(t)
    dfy = lambda t: 0.5 * hight * np.cos(t)
    t = np.linspace(1 / 6, 5 / 6, 5) * np.pi
    path, codes = make_path(fx, fy, dfx, dfy, t)

    pp1 = PathPatch(Path(path, codes),
                    ls=ls,
                    lw=lw,
                    fc=fc,
                    ec=ec,
                    transform=ax.transData)

    ax.add_patch(pp1)

    angle = np.pi / 6

    ax.arrow(x,
             y - 0.25 * width,
             0.55 * width * np.sin(angle),
             0.55 * width * np.cos(angle),
             head_width=0.07 * width,
             head_length=0.1 * width,
             lw=lw,
             color=ec)

    return (x - width / 2, x + width / 2)


def plot_gate(ax,
              x=0,
              y=0,
              size=1,
              text="$U$",
              ls='-',
              lw=2,
              r=0.2,
              fc='none',
              ec='black',
              fontsize=16):
    width = size
    hight = 0.8 * size
    plot_square(ax, x, y, width, hight, r, ls, lw, fc, ec)
    ax.text(x, y, text, ha='center', va='center', color=ec, fontsize=fontsize)

    return (x - width / 2, x + width / 2)


def plot_waveform(ax,
                  wav,
                  offset,
                  gaps=[],
                  ec='black',
                  fc='none',
                  ls='-',
                  lw=2):
    t = np.linspace(wav.start, wav.stop, 1001)
    points = wav(t) + offset
    for a, b in gaps:
        points[(t > a) * (t < b)] = np.nan
    ax.plot(t, points, ls=ls, color=ec, lw=lw)
    if fc != 'none':
        ax.fill_between(t, points, offset, color=fc)



from datetime import datetime

from sqlalchemy import (Column, DateTime, Float, ForeignKey, Integer, String,
                        Table, create_engine)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

Base = declarative_base()

dataset_tag_association = Table(
    'dataset_tag', Base.metadata,
    Column('dataset_id', Integer, ForeignKey('datasets.id'), primary_key=True),
    Column('tag_id', Integer, ForeignKey('tags.id'), primary_key=True))


class Experiment(Base):
    __tablename__ = 'experiments'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    start_date = Column(DateTime, default=datetime.utcnow)
    end_date = Column(DateTime, default=datetime.utcnow)
    researcher_id = Column(Integer, ForeignKey('researchers.id'))
    datasets = relationship('Dataset', back_populates='experiment')


class Researcher(Base):
    __tablename__ = 'researchers'
    id = Column(Integer, primary_key=True)
    first_name = Column(String)
    last_name = Column(String)
    email = Column(String)
    affiliation = Column(String)
    experiments = relationship('Experiment', back_populates='researcher')


class Dataset(Base):
    __tablename__ = 'datasets'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    tags = relationship('Tag',
                        secondary=dataset_tag_association,
                        back_populates='datasets')
    creation_date = Column(DateTime, default=datetime.utcnow)
    modification_date = Column(DateTime, default=datetime.utcnow)
    experiment_id = Column(Integer, ForeignKey('experiments.id'))
    experiment = relationship('Experiment', back_populates='datasets')
    data_entries = relationship('DataEntry', back_populates='dataset')


class DataEntry(Base):
    __tablename__ = 'data_entries'
    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    value = Column(Float)
    units = Column(String)
    variable_name = Column(String)
    dataset_id = Column(Integer, ForeignKey('datasets.id'))
    dataset = relationship('Dataset', back_populates='data_entries')


class Variable(Base):
    __tablename__ = 'variables'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    data_type = Column(String)


class Tag(Base):
    __tablename__ = 'tags'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    datasets = relationship('Dataset',
                            secondary=dataset_tag_association,
                            back_populates='tags')


# 创建数据库引擎
engine = create_engine('sqlite:///experiment_data.db')

# 创建表
Base.metadata.create_all(engine)

# 创建会话
Session = sessionmaker(bind=engine)
session = Session()



from multiprocessing import Queue
from typing import Literal

from .config import forms, ridx, style
from .dataset import Dataset
from .event_queue import EventQueue
from .ploter import PlotWidget
from .qt_compat import (BottomDockWidgetArea, QtCore, QtWidgets,
                        ScrollBarAlwaysOff, ScrollBarAlwaysOn,
                        TopDockWidgetArea)
from .toolbar import ToolBar


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self,
                 queue: Queue,
                 ncols=3,
                 plot_minimum_height=350,
                 plot_colors: list[tuple[int, int, int]] | None = None):
        super().__init__()
        self.ncols = ncols
        self.need_reshuffled = False
        self.plot_minimum_height = plot_minimum_height
        self.plot_widgets: list[PlotWidget] = []
        self.plot_colors = plot_colors
        self.toolbar = ToolBar()
        self.trace_data_box = Dataset()
        self.point_data_box = Dataset()
        self.queue = EventQueue(queue, self.toolbar, self.point_data_box,
                                self.trace_data_box)

        self.init_ui()

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(250)

    def init_ui(self):
        self.setStyleSheet(style)
        self.setMinimumHeight(500)
        self.setMinimumWidth(700)
        self.scroll = QtWidgets.QScrollArea(
        )  # Scroll Area which contains the widgets, set as the centralWidget
        self.widget = QtWidgets.QWidget(
        )  # Widget that contains the collection of Vertical Box
        self.layout = QtWidgets.QGridLayout()
        self.widget.setLayout(self.layout)

        #Scroll Area Properties
        #self.setCorner(Qt.TopSection, Qt.TopDockWidgetArea);
        self.scroll.setVerticalScrollBarPolicy(ScrollBarAlwaysOn)
        self.scroll.setHorizontalScrollBarPolicy(ScrollBarAlwaysOff)
        self.scroll.setWidgetResizable(True)
        self.scroll.setWidget(self.widget)
        self.setCentralWidget(self.scroll)

        self.dock = QtWidgets.QDockWidget(self)
        self.dock.setAllowedAreas(TopDockWidgetArea | BottomDockWidgetArea)
        self.addDockWidget(TopDockWidgetArea, self.dock)
        self.dock.setFloating(False)
        self.dock.setWidget(self.toolbar)
        self.dock.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable)
        #self.tabifyDockWidget(self.dock,None);
        #self.addDockWidget(self.dock);
        #self.setStatusBar(self.toolbar);
        #self.layout.addWidget(self.toolbar , 0 , 0  , 1, self.ncol);

        self.setWindowTitle('Scroll multi view')
        self.show()
        self.toolbar.set_mainwindow(self)
        self.toolbar.pb.setChecked(True)

    @property
    def mode(self) -> Literal["P", "T"]:
        return self.toolbar.mode

    @property
    def dataset(self) -> Dataset:
        return {"P": self.point_data_box, "T": self.trace_data_box}[self.mode]

    def set_ncols(self, x: int):
        x = max(1, min(10, int(x)))
        if (x != self.ncols):
            self.need_reshuffled = True
            self.ncols = x

    def add_subplot(self):
        n = len(self.plot_widgets)
        pw = PlotWidget(self.plot_minimum_height, self.plot_colors)
        self.plot_widgets.append(pw)
        grid_row = n // self.ncols
        grid_col = n % self.ncols
        self.layout.addWidget(pw, grid_row + 1, grid_col)
        return pw

    def create_subplots(self, xy_pairs):
        for xn, yn in xy_pairs:
            pw = self.add_subplot()
            pw.set_X_label(xn)
            pw.set_Y_label(yn)
        self.do_link()
        self.all_enable_auto_range()

    def clear_subplots(self):
        for i in range(len(self.plot_widgets)):
            self.layout.removeWidget(self.plot_widgets[i])
            self.plot_widgets[i].setParent(None)
        self.plot_widgets.clear()

    def remove_plot(self, w: PlotWidget):
        w.setParent(None)
        self.plot_widgets.remove(w)
        self.reshuffle()

    def drop_last_plot(self, i_=-1):
        # delete the one
        i = int(i_)
        if (i < len(self.plot_widgets)):
            w = self.plot_widgets[i]
            w.setParent(None)
            del w
            del self.plot_widgets[i]
        self.reshuffle()

    def reshuffle(self):
        for idx, widget in enumerate(self.plot_widgets):
            widget.setParent(None)
            grid_row = idx // self.ncols
            grid_col = idx % self.ncols
            self.layout.addWidget(widget, grid_row + 1, grid_col)

    def keyPressEvent(self, ev):
        #print(ev.text());
        tx = ev.text()
        if (tx in ['_', '-']):
            self.set_ncols(self.ncols - 1)
        elif (tx in ['=', '+']):
            self.set_ncols(self.ncols + 1)

    def mouse_click(self):
        pass

    def do_link(self):
        """
        link the plot

        share the same x or y axis
        """
        same_X = {}
        xy_pairs = self.toolbar.xypairs
        for idx, xyn in enumerate(xy_pairs):
            xn, yn = xyn
            if xn not in same_X:
                same_X[xn] = []
            same_X[xn].append(idx)

        sharex, sharey = self.toolbar.sharexy()

        s_A = not (sharex and sharey)
        for x, yidxs in same_X.items():
            pre_yidx = -1
            for yidx in yidxs:
                if (-1 != pre_yidx):
                    if (s_A):
                        self.plot_widgets[pre_yidx].plotItem.vb.setXLink(None)
                        self.plot_widgets[pre_yidx].plotItem.vb.setYLink(None)

                    if sharex:
                        self.plot_widgets[pre_yidx].plotItem.vb.setXLink(
                            self.plot_widgets[yidx].plotItem.vb)

                    if sharey:
                        self.plot_widgets[pre_yidx].plotItem.vb.setYLink(
                            self.plot_widgets[yidx].plotItem.vb)
                pre_yidx = yidx

    def all_auto_range(self):
        for pw in self.plot_widgets:
            pw.auto_range()

    def all_enable_auto_range(self):
        for pw in self.plot_widgets:
            pw.enable_auto_range()

    def update(self):
        # update the queue
        self.queue.flush()

        rescale = False

        # setup the xyfm
        if (self.toolbar.xypairs_dirty):
            self.clear_subplots()
            self.create_subplots(self.toolbar.xypairs)
            self.toolbar.xypairs_dirty = False
            rescale = True

        if (self.toolbar.link_dirty):
            self.do_link()
            self.toolbar.link_dirty = False

        if (self.need_reshuffled):
            self.need_reshuffled = False
            self.reshuffle()

        # checking the log space
        if (self.toolbar.xyfm_dirty):
            for pw in self.plot_widgets:
                pw.plotItem.ctrl.logXCheck.setChecked(self.toolbar.lx)
                pw.plotItem.ctrl.logYCheck.setChecked(self.toolbar.ly)

        #update the plot
        # if clear is set then do the clear :
        if (self.toolbar.CR_flag):
            self.toolbar.CR_flag = False
            self.dataset.clear_history()
            self.dataset.dirty = True

        if (self.dataset.dirty or self.toolbar.xyfm_dirty or rescale):
            self.dataset.dirty = False
            self.xyfm_dirty = False
            for pw in self.plot_widgets:

                fx = forms[self.toolbar.fx]
                fy = forms[self.toolbar.fy]
                for i in ridx:
                    x, y = self.dataset.get_data(i, pw.xname, pw.yname)
                    l = min(len(x), len(y))
                    x, y = fx(x[:l], 0), fy(y[:l], 0)
                    pw.set_data(i, x, y)
                pw.update()
                if rescale:
                    pw.auto_range()



import hashlib
import lzma
import pickle
import uuid
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from functools import lru_cache
from pathlib import Path
from typing import Any, Literal

from loguru import logger

from ..cli.config import get_config_value

__current_config_cache = None


@dataclass
class Result():
    workflow: str = ''
    in_spec: bool = False
    bad_data: bool = False
    fully_calibrated: bool = False
    calibrated_time: datetime = field(default_factory=datetime.now)
    checked_time: datetime = field(default_factory=datetime.now)
    ttl: timedelta = timedelta(days=3650)
    parameters: dict = field(default_factory=dict)
    oracle: dict = field(default_factory=dict)
    other_infomation: dict = field(default_factory=dict)
    data: Any = field(default_factory=tuple)
    index: int = -1
    previous_path: Path | None = None
    base_path: Path | None = None
    config_path: Path | None = None

    @property
    def previous(self):
        if self.previous_path is not None and self.base_path is not None:
            return load_result(self.previous_path, self.base_path)
        else:
            return None

    @property
    def state(self) -> Literal['OK', 'Bad', 'Outdated']:
        state = 'Bad'
        match (self.in_spec, self.bad_data):
            case (True, False):
                state = 'OK'
            case (False, True):
                state = 'Bad'
            case (False, False):
                state = 'Outdated'
        return state

    @property
    def config(self):
        if self.config_path is not None and self.base_path is not None:
            return load_config(self.config_path, self.base_path)
        else:
            return None

    @state.setter
    def state(self, state: Literal['OK', 'Bad', 'Outdated', 'In spec',
                                   'Out of spec', 'Bad data']):
        if state not in [
                'OK', 'Bad', 'Outdated', 'In spec', 'Out of spec', 'Bad data'
        ]:
            raise ValueError(
                f'Invalid state: {state}, state must be one of "OK", "Bad" and "Outdated"'
            )
        if state in ['In spec', 'OK']:
            self.in_spec = True
            self.bad_data = False
        elif state in ['Bad data', 'Bad']:
            self.bad_data = True
            self.in_spec = False
        else:
            self.bad_data = False
            self.in_spec = False


def random_path(base: Path) -> Path:
    while True:
        s = uuid.uuid4().hex
        path = Path(s[:2]) / s[2:4] / s[4:6] / s[6:]
        if not (base / path).exists():
            return path


def save_config_key_history(key: str, result: Result,
                            base_path: str | Path) -> int:
    global __current_config_cache
    base_path = Path(base_path) / 'state'
    base_path.mkdir(parents=True, exist_ok=True)

    if __current_config_cache is None:
        if (base_path / 'parameters.pkl').exists():
            with open(base_path / 'parameters.pkl', 'rb') as f:
                __current_config_cache = pickle.load(f)
        else:
            __current_config_cache = {}

    __current_config_cache[key] = result.data

    with open(base_path / 'parameters.pkl', 'wb') as f:
        pickle.dump(__current_config_cache, f)
    return 0


def find_config_key_history(key: str, base_path: str | Path) -> Result | None:
    global __current_config_cache
    base_path = Path(base_path) / 'state'
    if __current_config_cache is None:
        if (base_path / 'parameters.pkl').exists():
            with open(base_path / 'parameters.pkl', 'rb') as f:
                __current_config_cache = pickle.load(f)
        else:
            __current_config_cache = {}

    if key in __current_config_cache:
        value = __current_config_cache.get(key, None)
        result = Result(workflow=f'cfg:{key}',
                        bad_data=False,
                        in_spec=True,
                        fully_calibrated=True,
                        parameters={key: value},
                        data=value)
        result.bad_data = False
        return result
    return None


def save_result(workflow: str,
                result: Result,
                base_path: str | Path,
                overwrite: bool = False) -> int:
    if workflow.startswith("cfg:"):
        return save_config_key_history(workflow[4:], result, base_path)

    logger.debug(
        f'Saving result for "{workflow}", {result.in_spec=}, {result.bad_data=}, {result.fully_calibrated=}'
    )
    base_path = Path(base_path)
    if overwrite:
        buf = lzma.compress(pickle.dumps(result))
        path = get_head(workflow, base_path)
        with open(base_path / 'objects' / path, "rb") as f:
            index = int.from_bytes(f.read(8), 'big')
        result.index = index
    else:
        result.previous_path = get_head(workflow, base_path)
        buf = lzma.compress(pickle.dumps(result))
        path = random_path(base_path / 'objects')
        (base_path / 'objects' / path).parent.mkdir(parents=True,
                                                    exist_ok=True)
        result.index = create_index("result",
                                    base_path,
                                    context=str(path),
                                    width=35)

    with open(base_path / 'objects' / path, "wb") as f:
        f.write(result.index.to_bytes(8, 'big'))
        f.write(buf)
    set_head(workflow, path, base_path)
    return result.index


def load_result(path: str | Path, base_path: str | Path) -> Result | None:
    base_path = Path(base_path)
    path = base_path / 'objects' / path

    with open(base_path / 'objects' / path, "rb") as f:
        index = int.from_bytes(f.read(8), 'big')
        result = pickle.loads(lzma.decompress(f.read()))
        result.base_path = base_path
        result.index = index
        return result


def find_result(
    workflow: str, base_path: str | Path = get_config_value("data", Path)
) -> Result | None:
    if workflow.startswith("cfg:"):
        return find_config_key_history(workflow[4:], base_path)

    base_path = Path(base_path)
    path = get_head(workflow, base_path)
    if path is None:
        return None
    return load_result(path, base_path)


def renew_result(workflow: str, base_path: str | Path):
    logger.debug(f'Renewing result for "{workflow}"')
    result = find_result(workflow, base_path)
    if result is not None:
        result.checked_time = datetime.now()
        return save_result(workflow, result, base_path)


def revoke_result(workflow: str, base_path: str | Path):
    logger.debug(f'Revoking result for "{workflow}"')
    base_path = Path(base_path)
    path = get_head(workflow, base_path)
    if path is not None:
        result = load_result(path, base_path)
        result.in_spec = False
        return save_result(workflow, result, base_path)


def set_head(workflow: str, path: Path, base_path: str | Path):
    base_path = Path(base_path)
    base_path.mkdir(parents=True, exist_ok=True)
    try:
        with open(base_path / "heads", "rb") as f:
            heads = pickle.load(f)
    except:
        heads = {}
    heads[workflow] = path
    with open(base_path / "heads", "wb") as f:
        pickle.dump(heads, f)


def get_head(workflow: str, base_path: str | Path) -> Path | None:
    base_path = Path(base_path)
    try:
        with open(base_path / "heads", "rb") as f:
            heads = pickle.load(f)
        return heads[workflow]
    except:
        return None


def get_heads(base_path: str | Path) -> Path | None:
    base_path = Path(base_path)
    try:
        with open(base_path / "heads", "rb") as f:
            heads = pickle.load(f)
        return heads
    except:
        return {}


def create_index(name: str,
                 base_path: str | Path,
                 context: str,
                 width: int = -1,
                 start: int = 0):

    path = Path(base_path) / "index" / name
    if width == -1:
        width = len(context)
    else:
        width = max(width, len(context))

    if path.with_suffix('.width').exists():
        width = int(path.with_suffix('.width').read_text())
        index = int(path.with_suffix('.seq').read_text())
    else:
        index = start
    if width < len(context):
        raise ValueError(
            f"Context '{context}' is too long, existing width of '{name}' is {width}."
        )
    if not path.with_suffix('.width').exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.with_suffix('.width').write_text(str(width))

    path.with_suffix('.seq').write_text(str(index + 1))

    with path.with_suffix('.idx').open("a") as f:
        f.write(f"{context.ljust(width)}\n")

    return index


@lru_cache(maxsize=4096)
def query_index(name: str, base_path: str | Path, index: int):
    path = Path(base_path) / "index" / name
    width = int(path.with_suffix('.width').read_text())

    with path.with_suffix('.idx').open("r") as f:
        f.seek(index * (width + 1))
        context = f.read(width)
    return context.rstrip()


def get_result_by_index(
    index: int, base_path: str | Path = get_config_value("data", Path)
) -> Result | None:
    try:
        path = query_index("result", base_path, index)
        return load_result(path, base_path)
    except:
        return None


def save_config(cfg, data_path):
    i = 0
    buf = pickle.dumps(cfg)
    buf = lzma.compress(buf)
    h = hashlib.md5(buf)

    while True:
        salt = f"{i}".encode()
        h.update(salt)
        hashstr = h.hexdigest()
        cfg_id = Path(hashstr[:2]) / hashstr[2:4] / hashstr[4:]
        path = Path(data_path) / 'config' / cfg_id
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, 'wb') as f:
                f.write(buf)
            break
        elif path.read_bytes() == buf:
            break
        i += 1
    return str(cfg_id)


@lru_cache(maxsize=1024)
def load_config(id, data_path):
    path = Path(data_path) / 'config' / id
    with open(path, 'rb') as f:
        buf = f.read()
    cfg = pickle.loads(lzma.decompress(buf))
    return cfg



import asyncio
import pathlib
import pickle

import click

from .kad import Server


async def client_get(key, server, port, timeout=5):
    reader, writer = await asyncio.wait_for(asyncio.open_connection(
        server, port),
                                            timeout=timeout)
    writer.write(pickle.dumps(('get', key)))
    await asyncio.wait_for(writer.drain(), timeout=timeout)
    data = await asyncio.wait_for(reader.read(1600), timeout=timeout)
    value = pickle.loads(data)
    writer.close()
    await writer.wait_closed()
    return value


async def client_set(key, value, server, port, timeout=5):
    reader, writer = await asyncio.wait_for(asyncio.open_connection(
        server, port),
                                            timeout=timeout)
    writer.write(pickle.dumps(('set', key, value)))
    await asyncio.wait_for(writer.drain(), timeout=timeout)
    writer.close()
    await writer.wait_closed()


async def client_bootstrap(address, server, port, timeout=5):
    reader, writer = await asyncio.wait_for(asyncio.open_connection(
        server, port),
                                            timeout=timeout)
    addr, p = address.split(':')
    writer.write(pickle.dumps(('bootstrap', (addr, int(p)))))
    await asyncio.wait_for(writer.drain(), timeout=timeout)
    writer.close()
    await writer.wait_closed()


@click.group()
def dht():
    pass


@dht.command()
@click.option('--interface',
              default='127.0.0.1',
              help='Interface to listen on')
@click.option('--port', default=8467, help='Port to listen on')
@click.option('--dht-only', default=False, is_flag=True, help='DHT only')
@click.option('--dht-interface',
              default='0.0.0.0',
              help='DHT interface to listen on')
@click.option('--dht-port', default=8468, help='DHT port to listen on')
@click.option('--bootstrap', default=None, help='Address of bootstrap node')
@click.option('--interval', default=300, help='Refresh interval')
@click.option('--state-file', default=None, help='State file')
def server(interface, port, dht_only, dht_interface, dht_port, bootstrap,
           interval, state_file):

    async def main(interface, port, dht_only, dht_interface, dht_port,
                   bootstrap, state_file):

        if state_file is None:
            state_file = pathlib.Path.home() / '.waveforms' / 'dht.pickle'
        else:
            state_file = pathlib.Path(state_file)

        if state_file.exists():
            node = await Server.load_state(state_file, dht_port, dht_interface,
                                           interval)
        else:
            node = Server()
            state_file.parent.mkdir(parents=True, exist_ok=True)
            await node.listen(dht_port, dht_interface, interval)
        if bootstrap:
            addr, p = bootstrap.split(':')
            await node.bootstrap([(addr, int(p))])

        loop = asyncio.get_running_loop()
        loop.call_later(min(interval / 2, 30), node.save_state_regularly,
                        state_file, interval)

        async def handle_cmds(reader, writer):
            data = await reader.read(1600)
            cmd, *args = pickle.loads(data)
            match cmd:
                case 'get':
                    value = await node.get(*args)
                    data = pickle.dumps(value)
                case 'set':
                    value = await node.set(*args)
                    data = pickle.dumps(value)
                case 'ping':
                    data = pickle.dumps(node.node.id)
                case 'bootstrap':
                    await node.bootstrap(args)
                    data = pickle.dumps(None)
                case _:
                    data = pickle.dumps(None)

            writer.write(data)
            await writer.drain()
            writer.close()
            await writer.wait_closed()

        if dht_only:
            try:
                while True:
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                pass
        else:
            server = await asyncio.start_server(handle_cmds, interface, port)

            async with server:
                try:
                    await server.serve_forever()
                except asyncio.CancelledError:
                    pass

    asyncio.run(
        main(interface, port, dht_only, dht_interface, dht_port, bootstrap,
             state_file))


@dht.command('get')
@click.argument('key')
@click.option('--server', default='127.0.0.1', help='Server to connect to')
@click.option('--port', default=8467, help='Port to connect to')
@click.option('--timeout', default=5, help='Timeout')
def get(key, server, port, timeout):

    async def client():
        value = await client_get(key, server, port, timeout)
        print(value)

    asyncio.run(client())


@dht.command('set')
@click.argument('key')
@click.argument('value')
@click.option('--server', default='127.0.0.1', help='Server to connect to')
@click.option('--port', default=8467, help='Port to connect to')
@click.option('--timeout', default=5, help='Timeout')
def set_(key, value, server, port, timeout):

    async def client():
        await client_set(key, value, server, port, timeout)

    asyncio.run(client())


@dht.command('bootstrap')
@click.argument('address')
@click.option('--server', default='127.0.0.1', help='Server to connect to')
@click.option('--port', default=8467, help='Port to connect to')
@click.option('--timeout', default=5, help='Timeout')
def bootstrap(address, server, port, timeout):

    async def client():
        await client_bootstrap(address, server, port, timeout)

    asyncio.run(client())



import base64
import graphlib
import hashlib
import inspect
import lzma
import pickle
import re
import string
import textwrap
import warnings
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any

from loguru import logger

from .storage import Result


class SetConfigWorkflow():
    __timeout__ = None
    __mtime__ = 0

    def __init__(self, key):
        self.key = key
        self.__workflow_id__ = f"cfg:{self.key}"

    def depends(self):
        return []

    def check_state(self, history: Result) -> bool:
        from . import transform
        try:
            return self._equal(history.parameters[self.key],
                               transform.query_config(self.key))
        except:
            return False

    def calibrate(self):
        from . import transform
        try:
            value = transform.query_config(self.key)
        except:
            value = eval(input(f'"{self.key}": '))
        return value

    def analyze(self, result: Result, history: Result):
        result.state = 'OK'
        result.parameters = {self.key: result.data}
        return result

    def check(self):
        return self.calibrate()

    def check_analyze(self, result: Result, history: Result | None):
        if self.check_state(history):
            result.state = 'OK'
            result.parameters = {self.key: history.data}
        else:
            result.state = 'Outdated'
        return result

    @staticmethod
    def _equal(a, b):
        import numpy as np

        if a is b:
            return True
        try:
            return a == b
        except:
            pass

        if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
            return a.shape == b.shape and np.all(a == b)

        return False


WorkflowType = ModuleType | SetConfigWorkflow


def can_call_without_args(func):
    if not callable(func):
        return False

    # 获取函数签名
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        # 如果有参数没有默认值且不是可变参数，则无法无参调用
        if (param.default is param.empty and param.kind
                not in (param.VAR_POSITIONAL, param.VAR_KEYWORD)):
            return False
    return True


def verify_calibrate_method(module: WorkflowType):
    if not hasattr(module, 'calibrate'):
        raise AttributeError(
            f"Workflow {module.__file__} does not have 'calibrate' function")

    if not can_call_without_args(module.calibrate):
        raise AttributeError(
            f"Workflow {module.__file__} 'calibrate' function should not have any parameters"
        )

    if not hasattr(module, 'analyze'):
        raise AttributeError(
            f"Workflow {module.__file__} does not have 'analyze' function")


def verify_check_method(module: WorkflowType):
    if not hasattr(module, 'check'):
        warnings.warn(
            f"Workflow {module.__file__} does not have 'check' function, it will be set to 'calibrate' function"
        )
    else:
        if not can_call_without_args(module.check):
            raise AttributeError(
                f"Workflow {module.__file__} 'check' function should not have any parameters"
            )

        if not hasattr(module, 'check_analyze'):
            raise AttributeError(
                f"Workflow {module.__file__} has 'check' function but does not have 'check_analyze' function"
            )


def verify_dependence_key(workflow: str | tuple[str, dict[str, Any]]
                          | tuple[str, str, dict[str, Any]], base_path: Path):
    if isinstance(workflow, str):
        return
    if not isinstance(workflow, tuple) or len(workflow) not in [2, 3]:
        raise ValueError(f"Invalid workflow: {workflow}")

    if len(workflow) == 2:
        file_name, mapping = workflow
        if not Path(file_name).exists():
            raise FileNotFoundError(f"File not found: {file_name}")
    elif len(workflow) == 3:
        template_path, target_path, mapping = workflow
        if not (Path(base_path) / template_path).exists():
            raise FileNotFoundError(f"File not found: {template_path}")
        if not isinstance(target_path, (Path, str)) or target_path == '':
            raise ValueError(f"Invalid target_path: {target_path}")
        if not isinstance(target_path, (Path, str)):
            raise ValueError(f"Invalid target_path: {target_path}")
        if Path(target_path).suffix != '.py':
            raise ValueError(
                f"Invalid target_path: {target_path}. Only .py file is supported"
            )
    else:
        raise ValueError(f"Invalid workflow: {workflow}")

    if not isinstance(mapping, dict):
        raise ValueError(f"Invalid mapping: {mapping}")

    for key, value in mapping.items():
        if not isinstance(key, str):
            raise ValueError(
                f"Invalid key: {key}, should be str type and valid identifier")
        if not key.isidentifier():
            raise ValueError(f"Invalid key: {key}, should be identifier")
        try:
            pickle.dumps(value)
        except Exception as e:
            raise ValueError(
                f"Invalid value: {key}: {value}, should be pickleable") from e
    return


def verify_depends(module: WorkflowType, base_path):
    if not hasattr(module, 'depends'):
        return

    deps = []

    if callable(module.depends):
        if not can_call_without_args(module.depends):
            raise AttributeError(
                f"Workflow {module.__file__} 'depends' function should not have any parameters"
            )
        deps = list(module.depends())
    elif isinstance(module.depends, (list, tuple)):
        deps = module.depends
    else:
        raise AttributeError(
            f"Workflow {module.__file__} 'depends' should be a callable or a list"
        )
    for workflow in deps:
        verify_dependence_key(workflow, base_path)


def verify_entries(module: WorkflowType, base_path):
    if not hasattr(module, 'entries'):
        return

    deps = []

    if callable(module.entries):
        if not can_call_without_args(module.entries):
            raise AttributeError(
                f"Workflow {module.__file__} 'entries' function should not have any parameters"
            )
        deps = list(module.entries())
    elif isinstance(module.entries, (list, tuple)):
        deps = module.entries
    else:
        raise AttributeError(
            f"Workflow {module.__file__} 'entries' should be a callable or a list"
        )
    for workflow in deps:
        verify_dependence_key(workflow, base_path)


def is_workflow(module: ModuleType) -> bool:
    try:
        verify_calibrate_method(module)
        return True
    except AttributeError:
        return False


def is_template(path: str | Path) -> bool:
    path = Path(path)
    if path.name == 'template.py':
        return True
    if path.name.endswith('_template.py'):
        return True
    if 'templates' in path.parts:
        return True
    return False


def find_unreferenced_workflows(path: str) -> list[str]:
    root = Path(path).resolve()
    workflows = []
    workflow_paths: set[str] = set()

    # Collect all workflow modules
    for file_path in root.rglob("*.py"):
        if file_path.name == "__init__.py":
            continue
        if is_template(file_path):
            continue
        try:
            rel_path = file_path.relative_to(root)
        except ValueError:
            continue

        module = load_workflow_from_file(str(rel_path), root)

        if is_workflow(module):
            rel_str = str(rel_path)
            workflows.append(rel_str)
            workflow_paths.add(rel_str)

    dependencies: set[str] = set()

    # Check dependencies for each workflow module
    for rel_str in workflows:
        module = load_workflow_from_file(rel_str, root)

        depends_func = getattr(module, "depends", None)
        if depends_func and callable(depends_func):
            if not can_call_without_args(depends_func):
                warnings.warn(
                    f"Skipping depends() in {rel_str} as it requires arguments"
                )
                continue
            try:
                depends_list = [
                    n.__workflow_id__ for n in get_dependents(module, root)
                ]
            except Exception as e:
                warnings.warn(f"Error calling depends() in {rel_str}: {e}")
                continue

            if not isinstance(depends_list, list) or not all(
                    isinstance(item, str) for item in depends_list):
                warnings.warn(
                    f"depends() in {rel_str} did not return a list of strings")
                continue

            for dep in depends_list:
                dep_full = (root / dep).resolve()
                try:
                    dep_rel = dep_full.relative_to(root)
                except ValueError:
                    continue
                dep_rel_str = str(dep_rel)
                if dep_rel_str in workflow_paths:
                    dependencies.add(dep_rel_str)

    # Determine unreferenced workflows
    unreferenced = [wp for wp in workflows if wp not in dependencies]
    return unreferenced


def load_workflow_from_file(file_name: str,
                            base_path: str | Path,
                            package='workflows') -> WorkflowType:
    base_path = Path(base_path)
    path = Path(file_name)
    if not (base_path / path).exists():
        raise FileNotFoundError(f"File not found: {base_path / path}")
    module_name = f"{package}.{'.'.join([*path.parts[:-1], path.stem])}"
    spec = spec_from_file_location(module_name, base_path / path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    module.__mtime__ = (base_path / path).stat().st_mtime

    if hasattr(module, 'entries'):
        verify_entries(module, base_path)
        return module

    if not hasattr(module, '__timeout__'):
        module.__timeout__ = None

    if not hasattr(module, 'depends'):
        module.depends = lambda: []
    verify_depends(module, base_path)
    verify_calibrate_method(module)
    verify_check_method(module)

    return module


def encode_mapping(mapping):
    mapping_bytes = lzma.compress(pickle.dumps(mapping))
    hash_str = hashlib.md5(mapping_bytes).hexdigest()[:8]
    mappping_code = '\n'.join(
        textwrap.wrap(base64.b64encode(mapping_bytes).decode(), 100))
    return hash_str, mappping_code


def decode_mapping(hash_str, mappping_code):
    mapping_bytes = base64.b64decode(mappping_code.replace('\n', ''))
    if hash_str != hashlib.md5(mapping_bytes).hexdigest()[:8]:
        raise ValueError("Hash does not match")
    mapping = pickle.loads(lzma.decompress(mapping_bytes))
    return mapping


def load_workflow_from_template(template_path: str,
                                mapping: dict[str, str],
                                base_path: str | Path,
                                target_path: str | None = None,
                                package='workflows',
                                mtime: float = 0) -> WorkflowType:
    base_path = Path(base_path)
    path = Path(template_path)

    with open(base_path / path) as f:
        content = f.read()

    mtime = max((base_path / template_path).stat().st_mtime, mtime)
    hash_str, mapping_code = encode_mapping(mapping)

    def replace(text):
        """
        将给定文本中的所有 VAR("var") 替换为 __VAR_{hash_str}["var"]。
        
        Args:
            text (str): 包含 VAR 调用的字符串。
        
        Returns:
            str: 已经替换的新字符串。
        """
        pattern = re.compile(r'VAR\s*\(\s*(["\'])(\w+)\1\s*\)')
        replacement = f'__VAR_{hash_str}' + r'[\1\2\1]'
        new_text = re.sub(pattern, replacement, text)
        return new_text

    template = string.Template(replace(content))
    keys = template.get_identifiers()
    missing = set(keys) - set(mapping.keys())
    if missing:
        raise KeyError(f"{template_path}: Missing keys in mapping: {missing}")
    content = template.substitute(mapping)

    inject_code = [
        "from qulab.executor.load import decode_mapping",
        f"__VAR_{hash_str} = decode_mapping(\"{hash_str}\",",
        f"\"\"\"{mapping_code}\"\"\")"
    ]
    content = '\n'.join(inject_code + [content])
    if target_path is None:
        if path.stem == 'template':
            path = path.parent / f'tmp{hash_str}.py'
        elif path.stem.endswith('_template'):
            path = path.parent / path.stem.replace('_template',
                                                   f'_tmp{hash_str}.py')
        else:
            path = path.parent / f'{path.stem}_tmp{hash_str}.py'
    else:
        path = target_path

    file = base_path / path
    if not file.exists():
        file.parent.mkdir(parents=True, exist_ok=True)
        with open(file, 'w') as f:
            f.write(content)
    else:
        if file.stat().st_mtime < mtime:
            with open(file, 'w') as f:
                f.write(content)
        else:
            if file.read_text() != content:
                logger.warning(
                    f"`{file}` already exists and is different from the new one generated from template `{template_path}`"
                )

    module = load_workflow_from_file(str(path), base_path, package)
    module.__mtime__ = max(mtime, module.__mtime__)

    return module


def load_workflow(workflow: str | tuple[str, dict],
                  base_path: str | Path,
                  package='workflows',
                  mtime: float = 0) -> WorkflowType:
    if isinstance(workflow, tuple):
        if len(workflow) == 2:
            file_name, mapping = workflow
            w = load_workflow_from_template(file_name, mapping, base_path,
                                            None, package, mtime)
        elif len(workflow) == 3:
            template_path, target_path, mapping = workflow
            w = load_workflow_from_template(template_path, mapping, base_path,
                                            target_path, package, mtime)
        else:
            raise ValueError(f"Invalid workflow: {workflow}")
        w.__workflow_id__ = str(Path(w.__file__).relative_to(base_path))
    elif isinstance(workflow, str):
        if workflow.startswith('cfg:'):
            key = workflow[4:]
            w = SetConfigWorkflow(key)
            w.__workflow_id__ = workflow
        else:
            w = load_workflow_from_file(workflow, base_path, package)
            w.__workflow_id__ = str(Path(w.__file__).relative_to(base_path))
    else:
        raise TypeError(f"Invalid workflow: {workflow}")

    return w


def get_dependents(workflow: WorkflowType,
                   code_path: str | Path) -> list[WorkflowType]:
    if callable(getattr(workflow, 'depends', None)):
        if not can_call_without_args(workflow.depends):
            raise AttributeError(
                f'Workflow {workflow.__workflow_id__} "depends" function should not have any parameters'
            )
        return [
            load_workflow(n, code_path, mtime=workflow.__mtime__)
            for n in workflow.depends()
        ]
    elif isinstance(getattr(workflow, 'depends', None), (list, tuple)):
        return [
            load_workflow(n, code_path, mtime=workflow.__mtime__)
            for n in workflow.depends
        ]
    elif getattr(workflow, 'entries', None) is None:
        return []
    else:
        raise AttributeError(
            f'Workflow {workflow.__workflow_id__} "depends" should be a callable or a list'
        )


def get_entries(workflow: WorkflowType,
                code_path: str | Path) -> list[WorkflowType]:
    if callable(getattr(workflow, 'entries', None)):
        if not can_call_without_args(workflow.entries):
            raise AttributeError(
                f'Workflow {workflow.__workflow_id__} "entries" function should not have any parameters'
            )
        return [
            load_workflow(n, code_path, mtime=workflow.__mtime__)
            for n in workflow.entries()
        ]
    elif isinstance(getattr(workflow, 'entries', None), (list, tuple)):
        return [
            load_workflow(n, code_path, mtime=workflow.__mtime__)
            for n in workflow.entries
        ]
    elif getattr(workflow, 'entries', None) is None:
        return []
    else:
        raise AttributeError(
            f'Workflow {workflow.__workflow_id__} "entries" should be a callable or a list'
        )


def make_graph(workflow: WorkflowType, graph: dict, code_path: str | Path):
    if workflow.__workflow_id__ in graph:
        raise graphlib.CycleError(
            f"Workflow {workflow.__workflow_id__} has a circular dependency")
    graph[workflow.__workflow_id__] = []

    if hasattr(workflow, 'entries'):
        for w in get_entries(workflow, code_path):
            graph[workflow.__workflow_id__].append(w.__workflow_id__)
            make_graph(w, graph=graph, code_path=code_path)
    elif hasattr(workflow, 'depends'):
        for w in get_dependents(workflow, code_path):
            graph[workflow.__workflow_id__].append(w.__workflow_id__)
            make_graph(w, graph=graph, code_path=code_path)
    if graph[workflow.__workflow_id__] == []:
        del graph[workflow.__workflow_id__]

    return graph



import struct

import numpy as np

types = {
    "b"      : (  int, 'b'), "B"      : (  int, 'B'),
    "h"      : (  int, 'h'), "H"      : (  int, 'H'),
    "i"      : (  int, 'i'), "I"      : (  int, 'I'),
    "q"      : (  int, 'q'), "Q"      : (  int, 'Q'),
    "f"      : (float, 'f'), "d"      : (float, 'd'),
    "int8"   : (  int, 'b'), "uint8"  : (  int, 'B'),
    "int16"  : (  int, 'h'), "uint16" : (  int, 'H'),
    "int32"  : (  int, 'i'), "uint32" : (  int, 'I'),
    "int64"  : (  int, 'q'), "uint64" : (  int, 'Q'),
    "float"  : (float, 'f'), "double" : (float, 'd'),
    "float32": (float, 'f'), "float64": (float, 'd')
} # yapf: disable


def IEEE_488_2_BinBlock(datalist, dtype="int16", is_big_endian=True):
    """
    将一组数据打包成 IEEE 488.2 标准二进制块

    Args:
        datalist : 要打包的数字列表
        dtype    : 数据类型
        endian   : 字节序

    Returns:
        binblock, header
        二进制块, 以及其 'header'
    """
    if isinstance(datalist, bytes):
        datablock = datalist
    else:
        datalist = np.asarray(datalist)
        datalist.astype(types[dtype][0])
        if is_big_endian:
            endianc = '>'
        else:
            endianc = '<'
        datablock = struct.pack(
            '%s%d%s' % (endianc, len(datalist), types[dtype][1]), *datalist)
    size = '%d' % len(datablock)
    header = '#%d%s' % (len(size), size)

    return header.encode() + datablock


def decode_IEEE_488_2_BinBlock(data, dtype="int16", is_big_endian=True):
    """
    解析 IEEE 488.2 标准二进制块

    Args:
        data : 二进制块
        dtype : 数据类型
        endian : 字节序

    Returns:
        datalist
    """
    if isinstance(data, bytes):
        header = data[:2]
        size = int(data[2:2 + int(header[1:2])])
        assert size == len(data) - 2 - int(header[1:2]), "data size error"
        datablock = data[2 + int(header[1:2]):size + 2 + int(header[1:2])]
    else:
        raise ValueError("data must be bytes")

    if is_big_endian:
        endianc = '>'
    else:
        endianc = '<'
    datalist = list(
        struct.unpack(
            '%s%d%s' % (endianc, size // struct.calcsize(types[dtype][1]),
                        types[dtype][1]), datablock))

    return datalist



import contextlib
import time
from datetime import datetime

from .storage.crud import create_cell, create_notebook
from .storage.models import Notebook

__current_notebook_id = None


def _get_current_notebook(session):
    global __current_notebook_id
    if __current_notebook_id is None:
        notebook = create_notebook(session, 'Untitled')
        session.commit()
        __current_notebook_id = notebook.id
    else:
        notebook = session.query(Notebook).filter(
            Notebook.id == __current_notebook_id).one()
    return notebook


def _get_current_cell(session):
    notebook = _get_current_notebook(session)
    if notebook.cells:
        return notebook.cells[-1]
    else:
        return None


def _save_inputCells(session, inputCells=None):
    notebook = _get_current_notebook(session)
    aready_saved = len(notebook.cells)
    if inputCells:
        for cell in inputCells[aready_saved:]:
            create_cell(session, notebook, cell)


def _update_inputCells(session):
    cell = _get_current_cell(session)
    if cell:
        cell.ftime = datetime.utcnow()
        session.add(cell)


__sessionmaker = None


@contextlib.contextmanager
def get_session():
    global __session
    if __sessionmaker is None:
        raise RuntimeError('Sessionmaker is not set')
    session = __sessionmaker()
    try:
        yield session
        session.commit()
    except:
        session.rollback()


def set_sessionmaker(sessionmaker):
    global __sessionmaker
    __sessionmaker = sessionmaker


def get_current_notebook(session=None):
    try:
        if session is None:
            with get_session() as session:
                return _get_current_notebook(session)
        else:
            return _get_current_notebook(session)
    except RuntimeError:
        return None


def get_current_cell_id(session=None):
    try:
        if session is None:
            with get_session() as session:
                return _get_current_cell(session).id
        else:
            return _get_current_cell(session).id
    except RuntimeError as e:
        return None


def get_inputCells():
    try:
        from IPython import get_ipython

        return get_ipython().user_ns['In']
    except Exception as e:
        # raise e
        return ['']


def autosave_cells(ipython):
    try:
        with get_session() as session:
            _save_inputCells(session, get_inputCells())
    except:
        pass


def update_timestamp(ipython):
    try:
        with get_session() as session:
            _update_inputCells(session)
    except:
        pass


def setup_ipy_events():
    try:
        from IPython import get_ipython

        ipython = get_ipython()

        if ipython is not None:
            ipython.events.register('pre_run_cell', autosave_cells)
            ipython.events.register('post_run_cell', update_timestamp)
    except:
        pass
